@@ -3,6 +3,7 @@
 #
 
 import re
+from Cython import Utils
 from Errors import warning, error, InternalError
 import Options
 import Naming
@@ -15,7 +16,8 @@ from TypeSlots import \
 import ControlFlow
 import __builtin__
 
-identifier_pattern = re.compile(r"[A-Za-z_][A-Za-z0-9_]*$")
+possible_identifier = re.compile(ur"(?![0-9])\w+$", re.U).match
+nice_identifier = re.compile('^[a-zA-Z0-0_]+$').match
 
 class Entry:
     # A symbol table entry in a Scope or ModuleNamespace.
@@ -64,9 +66,9 @@ class Entry:
     #                                 type is an extension type
     # as_module        None       Module scope, if a cimported module
     # is_inherited     boolean    Is an inherited attribute of an extension type
-    # #interned_cname   string     C name of interned name string
     # pystring_cname   string     C name of Python version of string literal
     # is_interned      boolean    For string const entries, value is interned
+    # is_identifier    boolean    For string const entries, value is an identifier
     # used             boolean
     # is_special       boolean    Is a special method or property accessor
     #                               of an extension type
@@ -104,8 +106,8 @@ class Entry:
     in_cinclude = 0
     as_module = None
     is_inherited = 0
-    #interned_cname = None
     pystring_cname = None
+    is_identifier = 0
     is_interned = 0
     used = 0
     is_special = 0
@@ -121,6 +123,9 @@ class Entry:
         self.pos = pos
         self.init = init
         
+    def redeclared(self, pos):
+        error(pos, "'%s' does not match previous declaration" % self.name)
+        error(self.pos, "Previous declaration is here")
         
 class Scope:
     # name              string             Unqualified name
@@ -149,12 +154,14 @@ class Scope:
     # pystring_entries  [Entry]            String const entries newly used as
     #                                        Python strings in this scope
     # control_flow     ControlFlow  Used for keeping track of environment state
+    # nogil             boolean            In a nogil section
 
     is_py_class_scope = 0
     is_c_class_scope = 0
     is_module_scope = 0
     scope_prefix = ""
     in_cinclude = 0
+    nogil = 0
     
     def __init__(self, name, outer_scope, parent_scope):
         # The outer_scope is the next scope in the lookup chain.
@@ -188,6 +195,7 @@ class Scope:
         self.cname_to_entry = {}
         self.pow_function_used = 0
         self.string_to_entry = {}
+        self.identifier_to_entry = {}
         self.num_to_entry = {}
         self.obj_to_entry = {}
         self.pystring_entries = []
@@ -204,10 +212,10 @@ class Scope:
         
     def __str__(self):
         return "<%s %s>" % (self.__class__.__name__, self.qualified_name)
-    
-    def intern(self, name):
-        return self.global_scope().intern(name)
-    
+
+    def intern_identifier(self, name):
+        return self.global_scope().intern_identifier(name)
+
     def qualifying_scope(self):
         return self.parent_scope
     
@@ -305,7 +313,8 @@ class Scope:
                 visibility = visibility, defining = scope is not None)
             self.sue_entries.append(entry)
         else:
-            if not (entry.is_type and entry.type.is_struct_or_union):
+            if not (entry.is_type and entry.type.is_struct_or_union
+                    and entry.type.kind == kind):
                 warning(pos, "'%s' redeclared  " % name, 0)
             elif scope and entry.type.scope:
                 warning(pos, "'%s' already defined  (ignoring second definition)" % name, 0)
@@ -365,6 +374,10 @@ class Scope:
     
     def declare_pyfunction(self, name, pos):
         # Add an entry for a Python function.
+        entry = self.lookup_here(name)
+        if entry and not entry.type.is_cfunction:
+            # This is legal Python, but for now will produce invalid C.
+            error(pos, "'%s' already declared" % name)
         entry = self.declare_var(name, py_object_type, pos)
         entry.signature = pyfunction_signature
         self.pyfunc_entries.append(entry)
@@ -415,6 +428,21 @@ class Scope:
         else:
             error(pos, "'%s' is not declared" % name)
     
+    def find_imported_module(self, path, pos):
+        # Look up qualified name, must be a module, report error if not found.
+        # Path is a list of names.
+        scope = self
+        for name in path:
+            entry = scope.find(name, pos)
+            if not entry:
+                return None
+            if entry.as_module:
+                scope = entry.as_module
+            else:
+                error(pos, "'%s' is not a cimported module" % scope.qualified_name)
+                return None
+        return scope
+        
     def lookup(self, name):
         # Look up name in this scope or an enclosing one.
         # Return None if not found.
@@ -434,9 +462,12 @@ class Scope:
             entry = self.declare_var(name, py_object_type, None)
         return entry
 
-    def add_string_const(self, value):
+    def add_string_const(self, value, identifier = False):
         # Add an entry for a string constant.
-        cname = self.new_const_cname()
+        if identifier:
+            cname = self.new_string_const_cname(value)
+        else:
+            cname = self.new_const_cname()
         if value.is_unicode:
             c_type = PyrexTypes.c_utf8_char_array_type
             value = value.utf8encode()
@@ -448,31 +479,36 @@ class Scope:
         self.const_entries.append(entry)
         return entry
 
-    def get_string_const(self, value):
+    def get_string_const(self, value, identifier = False):
         # Get entry for string constant. Returns an existing
         # one if possible, otherwise creates a new one.
         genv = self.global_scope()
-        entry = genv.string_to_entry.get(value)
+        if identifier:
+            string_map = genv.identifier_to_entry
+        else:
+            string_map = genv.string_to_entry
+        entry = string_map.get(value)
         if not entry:
-            entry = self.add_string_const(value)
-            genv.string_to_entry[value] = entry
+            entry = self.add_string_const(value, identifier)
+            entry.is_identifier = identifier
+            string_map[value] = entry
         return entry
 
-    def add_py_string(self, entry):
+    def add_py_string(self, entry, identifier = None):
         # If not already done, allocate a C name for a Python version of
         # a string literal, and add it to the list of Python strings to
         # be created at module init time. If the string resembles a
         # Python identifier, it will be interned.
-        if not entry.pystring_cname:
-            value = entry.init
-            if not entry.type.is_unicode and identifier_pattern.match(value):
-                entry.pystring_cname = self.intern(value)
-                entry.is_interned = 1
-            else:
-                entry.pystring_cname = entry.cname + "p"
-                self.pystring_entries.append(entry)
-                self.global_scope().all_pystring_entries.append(entry)
-                
+        if entry.pystring_cname:
+            return
+        value = entry.init
+        entry.pystring_cname = Naming.py_const_prefix + entry.cname[len(Naming.const_prefix):]
+        self.pystring_entries.append(entry)
+        self.global_scope().all_pystring_entries.append(entry)
+        if identifier or (identifier is None and possible_identifier(value)):
+            entry.is_interned = 1
+            self.global_scope().new_interned_string_entries.append(entry)
+
     def add_py_num(self, value):
         # Add an entry for an int constant.
         cname = "%s%s" % (Naming.interned_num_prefix, value)
@@ -505,6 +541,13 @@ class Scope:
             genv.obj_to_entry[obj] = entry
         return entry
     
+    def new_string_const_cname(self, value):
+        # Create a new globally-unique nice name for a string constant.
+        if len(value) < 20 and nice_identifier(value):
+            return "%s%s" % (Naming.const_prefix, value)
+        else:
+            return self.global_scope().new_const_cname()
+
     def new_const_cname(self):
         # Create a new globally-unique name for a constant.
         return self.global_scope().new_const_cname()
@@ -547,10 +590,6 @@ class Scope:
         return [entry for entry in self.temp_entries
             if entry not in self.free_temp_entries]
     
-    #def recycle_pending_temps(self):
-    #	# Obsolete
-    #	pass
-
     def use_utility_code(self, new_code):
         self.global_scope().use_utility_code(new_code)
     
@@ -605,11 +644,14 @@ class BuiltinScope(Scope):
             utility_code = None):
         # If python_equiv == "*", the Python equivalent has the same name
         # as the entry, otherwise it has the name specified by python_equiv.
+        name = Utils.EncodedString(name)
         entry = self.declare_cfunction(name, type, None, cname)
         entry.utility_code = utility_code
         if python_equiv:
             if python_equiv == "*":
                 python_equiv = name
+            else:
+                python_equiv = Utils.EncodedString(python_equiv)
             var_entry = Entry(python_equiv, python_equiv, py_object_type)
             var_entry.is_variable = 1
             var_entry.is_builtin = 1
@@ -617,6 +659,7 @@ class BuiltinScope(Scope):
         return entry
         
     def declare_builtin_type(self, name, cname):
+        name = Utils.EncodedString(name)
         type = PyrexTypes.BuiltinObjectType(name, cname)
         type.set_scope(CClassScope(name, outer_scope=None, visibility='extern'))
         self.type_names[name] = 1
@@ -641,7 +684,7 @@ class BuiltinScope(Scope):
         "long":   ["((PyObject*)&PyLong_Type)", py_object_type],
         "float":  ["((PyObject*)&PyFloat_Type)", py_object_type],
         
-        "str":    ["((PyObject*)&PyString_Type)", py_object_type],
+        "str":    ["((PyObject*)&PyBytes_Type)", py_object_type],
         "unicode":["((PyObject*)&PyUnicode_Type)", py_object_type],
         "tuple":  ["((PyObject*)&PyTuple_Type)", py_object_type],
         "list":   ["((PyObject*)&PyList_Type)", py_object_type],
@@ -671,19 +714,22 @@ class ModuleScope(Scope):
     # python_include_files [string]           Standard  Python headers to be included
     # include_files        [string]           Other C headers to be included
     # string_to_entry      {string : Entry}   Map string const to entry
+    # identifier_to_entry  {string : Entry}   Map identifier string const to entry
     # context              Context
     # parent_module        Scope              Parent in the import namespace
     # module_entries       {string : Entry}   For cimport statements
     # type_names           {string : 1}       Set of type names (used during parsing)
+    # included_files       [string]           Cython sources included with 'include'
     # pxd_file_loaded      boolean            Corresponding .pxd file has been processed
     # cimported_modules    [ModuleScope]      Modules imported with cimport
-    # intern_map           {string : string}  Mapping from Python names to interned strs
-    # interned_names       [string]           Interned names pending generation of declarations
+    # new_interned_string_entries [Entry]     New interned strings waiting to be declared
     # interned_nums        [int/long]         Interned numeric constants
     # all_pystring_entries [Entry]            Python string consts from all scopes
     # types_imported       {PyrexType : 1}    Set of types for which import code generated
+    # has_import_star      boolean            Module contains import *
     
     is_module_scope = 1
+    has_import_star = 0
 
     def __init__(self, name, parent_module, context):
         self.parent_module = parent_module
@@ -705,12 +751,12 @@ class ModuleScope(Scope):
         self.type_names = dict(outer_scope.type_names)
         self.pxd_file_loaded = 0
         self.cimported_modules = []
-        self.intern_map = {}
-        self.interned_names = []
+        self.new_interned_string_entries = []
         self.interned_nums = []
         self.interned_objs = []
         self.all_pystring_entries = []
         self.types_imported = {}
+        self.included_files = []
         self.pynum_entries = []
         self.has_extern_class = 0
         self.cached_builtins = []
@@ -724,7 +770,10 @@ class ModuleScope(Scope):
     
     def declare_builtin(self, name, pos):
         if not hasattr(__builtin__, name):
-            if self.outer_scope is not None:
+            if self.has_import_star:
+                entry = self.declare_var(name, py_object_type, pos)
+                return entry
+            elif self.outer_scope is not None:
                 return self.outer_scope.declare_builtin(name, pos)
             else:
                 error(pos, "undeclared name not builtin: %s"%name)
@@ -743,15 +792,11 @@ class ModuleScope(Scope):
         else:
             entry.is_builtin = 1
         return entry
-    
-    def intern(self, name):
-        intern_map = self.intern_map
-        cname = intern_map.get(name)
-        if not cname:
-            cname = Naming.interned_prefix + name
-            intern_map[name] = cname
-            self.interned_names.append(name)
-        return cname
+
+    def intern_identifier(self, name):
+        string_entry = self.get_string_const(name, identifier = True)
+        self.add_py_string(string_entry, identifier = 1)
+        return string_entry.pystring_cname
 
     def find_module(self, module_name, pos):
         # Find a module in the import namespace, interpreting
@@ -832,8 +877,6 @@ class ModuleScope(Scope):
                     "Non-cdef global variable is not a generic Python object")
             entry.is_pyglobal = 1
             entry.namespace_cname = self.module_cname
-            #if Options.intern_names:
-            #	entry.interned_cname = self.intern(name)
         else:
             entry.is_cglobal = 1
             self.var_entries.append(entry)
@@ -857,7 +900,7 @@ class ModuleScope(Scope):
         prefix=''
         n = self.const_counter
         self.const_counter = n + 1
-        return "%s%s_%d" % (Naming.const_prefix, prefix, n)
+        return "%s%s%d" % (Naming.const_prefix, prefix, n)
     
     def use_utility_code(self, new_code):
         #  Add string to list of utility code to be included,
@@ -867,23 +910,28 @@ class ModuleScope(Scope):
                 return
         self.utility_code_used.append(new_code)
     
-    def declare_c_class(self, name, pos, defining, implementing,
-        module_name, base_type, objstruct_cname, typeobj_cname,
-        visibility, typedef_flag, api):
+    def declare_c_class(self, name, pos, defining = 0, implementing = 0,
+        module_name = None, base_type = None, objstruct_cname = None,
+        typeobj_cname = None, visibility = 'private', typedef_flag = 0, api = 0):
         #
-        # Look for previous declaration as a type
+        #  Look for previous declaration as a type
         #
         entry = self.lookup_here(name)
         if entry:
             type = entry.type
             if not (entry.is_type and type.is_extension_type):
-                entry = None # Will cause an error when we redeclare it
+                entry = None # Will cause redeclaration and produce an error
             else:
-                self.check_previous_typedef_flag(entry, typedef_flag, pos)
-                if base_type != type.base_type:
-                    error(pos, "Base type does not match previous declaration")
+                scope = type.scope
+                if typedef_flag and (not scope or scope.defined):
+                    self.check_previous_typedef_flag(entry, typedef_flag, pos)
+                if (scope and scope.defined) or (base_type and type.base_type):
+                    if base_type and base_type is not type.base_type:
+                        error(pos, "Base type does not match previous declaration")
+                if base_type and not type.base_type:
+                    type.base_type = base_type
         #
-        # Make a new entry if needed
+        #  Make a new entry if needed
         #
         if not entry:
             type = PyrexTypes.PyExtensionType(name, typedef_flag, base_type)
@@ -905,7 +953,7 @@ class ModuleScope(Scope):
             self.attach_var_entry_to_c_class(entry)
             self.c_class_entries.append(entry)
         #
-        # Check for re-definition and create scope if needed
+        #  Check for re-definition and create scope if needed
         #
         if not type.scope:
             if defining or implementing:
@@ -923,7 +971,7 @@ class ModuleScope(Scope):
             elif implementing and type.scope.implemented:
                 error(pos, "C class '%s' already implemented" % name)
         #
-        # Fill in options, checking for compatibility with any previous declaration
+        #  Fill in options, checking for compatibility with any previous declaration
         #
         if defining:
             entry.defined_in_pxd = 1
@@ -1114,8 +1162,8 @@ class ClassScope(Scope):
         self.class_name = name
         self.doc = None
 
-    def add_string_const(self, value):
-        return self.outer_scope.add_string_const(value)
+    def add_string_const(self, value, identifier = False):
+        return self.outer_scope.add_string_const(value, identifier)
 
     def lookup(self, name):
         if name == "classmethod":
@@ -1151,8 +1199,6 @@ class PyClassScope(ClassScope):
             cname, visibility, is_cdef)
         entry.is_pyglobal = 1
         entry.namespace_cname = self.class_obj_cname
-        #if Options.intern_names:
-        #	entry.interned_cname = self.intern(name)
         return entry
 
     def allocate_temp(self, type):
@@ -1250,10 +1296,9 @@ class CClassScope(ClassScope):
                                   # I keep it in for now. is_member should be enough
                                   # later on
             entry.namespace_cname = "(PyObject *)%s" % self.parent_type.typeptr_cname
-            if Options.intern_names:
-                entry.interned_cname = self.intern(name)
+            entry.interned_cname = self.intern_identifier(name)
             return entry
-            
+
 
     def declare_pyfunction(self, name, pos):
         # Add an entry for a method.
@@ -1262,7 +1307,7 @@ class CClassScope(ClassScope):
         if name == "__new__":
             warning(pos, "__new__ method of extension type will change semantics "
                 "in a future version of Pyrex and Cython. Use __cinit__ instead.")
-            name = "__cinit__"
+            name = Utils.EncodedString("__cinit__")
         entry = self.declare_var(name, py_object_type, pos)
         special_sig = get_special_method_signature(name)
         if special_sig:
@@ -1279,7 +1324,7 @@ class CClassScope(ClassScope):
     
     def lookup_here(self, name):
         if name == "__new__":
-            name = "__cinit__"
+            name = Utils.EncodedString("__cinit__")
         return ClassScope.lookup_here(self, name)
     
     def declare_cfunction(self, name, type, pos,
@@ -1299,9 +1344,9 @@ class CClassScope(ClassScope):
                 if defining and entry.func_cname:
                     error(pos, "'%s' already defined" % name)
                 #print "CClassScope.declare_cfunction: checking signature" ###
-                if type.same_c_signature_as(entry.type, as_cmethod = 1):
+                if type.same_c_signature_as(entry.type, as_cmethod = 1) and type.nogil == entry.type.nogil:
                     pass
-                elif type.compatible_signature_with(entry.type, as_cmethod = 1):
+                elif type.compatible_signature_with(entry.type, as_cmethod = 1) and type.nogil == entry.type.nogil:
                     if type.optional_arg_count and not type.original_sig.optional_arg_count:
                         # Need to put a wrapper taking no optional arguments 
                         # into the method table.
@@ -1320,6 +1365,7 @@ class CClassScope(ClassScope):
 #                    entry.type = type
                 else:
                     error(pos, "Signature not compatible with previous declaration")
+                    error(entry.pos, "Previous declaration is here")
         else:
             if self.defined:
                 error(pos,
@@ -1359,8 +1405,8 @@ class CClassScope(ClassScope):
                 entry.is_variable = 1
                 self.inherited_var_entries.append(entry)
         for base_entry in base_scope.cfunc_entries:
-            entry = self.add_cfunction(base_entry.name, base_entry.type, None,
-                adapt(base_entry.cname), base_entry.visibility)
+            entry = self.add_cfunction(base_entry.name, base_entry.type,
+                    base_entry.pos, adapt(base_entry.cname), base_entry.visibility)
             entry.is_inherited = 1
             
     def allocate_temp(self, type):
@@ -1401,13 +1447,16 @@ static PyObject* __Pyx_Method_ClassMethod(PyObject *method); /*proto*/
 static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
     /* It appears that PyMethodDescr_Type is not anywhere exposed in the Python/C API */
     /* if (!PyObject_TypeCheck(method, &PyMethodDescr_Type)) { */ 
-    if (strcmp(method->ob_type->tp_name, "method_descriptor") == 0) { /* cdef classes */
+    if (strcmp(Py_TYPE(method)->tp_name, "method_descriptor") == 0) { /* cdef classes */
         PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
         return PyDescr_NewClassMethod(descr->d_type, descr->d_method);
     }
     else if (PyMethod_Check(method)) {                                /* python classes */
         return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
     }
+    else if (PyCFunction_Check(method)) {
+        return PyClassMethod_New(method);
+    }
     PyErr_Format(PyExc_TypeError, "Class-level classmethod() can only be called on a method_descriptor or instance method.");
     return NULL;
 }
